import argparse
import atexit
import sys

from praetor.praetor import CallTracer


def get_output_directory():
    parser = argparse.ArgumentParser(description='Custom praetor settings.')
    parser.add_argument('--praetor-output',required=False, help='Directory for praetor output')
    args = parser.parse_args()
    try:
        output_directory = args.praetor_output
    except AttributeError:
        output_directory = './output'
    return output_directory


out_dir = get_output_directory()
tracer = CallTracer(output_directory=out_dir, only_main=True)
sys.setprofile(tracer)

atexit.register(tracer.close)


import psutil
import time
import os
import threading


class DynamicProcessMonitor:
    def __init__(self, base_interval=1.0):
        self.parent_pid = os.getpid()
        self.base_interval = base_interval
        self.running = False
        self.monitor_thread = None
        self.stats = []
        self.file_history = set()
        self._lock = threading.Lock()

    def start(self):
        """Start continuous base monitoring"""
        if self.running:
            return
        self.running = True

        def monitor_loop():
            while self.running:
                with self._lock:
                    self._snapshot('base')
                time.sleep(self.base_interval)

        self.monitor_thread = threading.Thread(target=monitor_loop, daemon=True)
        self.monitor_thread.start()

    def high_freq_snapshot(self, function_name):
        """Take IMMEDIATE high-frequency snapshot during monitoring"""
        with self._lock:
            return self._snapshot('high_freq', function_name)

    def _snapshot(self, snapshot_type, function_name):
        """Internal snapshot method"""
        snapshot = {
            'timestamp': time.time(),
            'type': snapshot_type,
            'processes': {}
        }

        newly_opened_files = []

        # Parent process
        try:
            parent_proc = psutil.Process(self.parent_pid)
            open_files_list = parent_proc.open_files()

            snapshot['processes'][self.parent_pid] = {
                'rss_mb': parent_proc.memory_info().rss / 1024 / 1024,
                'open_files_count': len(open_files_list)
            }

            # Only track NEWLY opened files
            current_files = {f.path for f in open_files_list}
            newly_opened_files = sorted(current_files - self.file_history)
            if snapshot_type == 'high_freq':
                self.file_history.update(current_files)

        except psutil.NoSuchProcess:
            pass

        # Child processes (track new files too)
        try:
            for child in psutil.Process().children(recursive=True):
                child_open_files = child.open_files()
                snapshot['processes'][child.pid] = {
                    'rss_mb': child.memory_info().rss / 1024 / 1024,
                    'open_files_count': len(child_open_files)
                }
        except psutil.NoSuchProcess:
            pass

        # NEWLY opened files list (across all processes)
        snapshot['newly_opened_files'] = newly_opened_files

        snapshot['total_rss_mb'] = sum(p['rss_mb'] for p in snapshot['processes'].values())
        snapshot['process_count'] = len(snapshot['processes'])
        snapshot['function_name'] = function_name
        self.stats.append(snapshot)
        return snapshot

    def stop(self):
        """Stop continuous monitoring"""
        self.running = False
        if self.monitor_thread:
            self.monitor_thread.join(timeout=2.0)

    def get_stats(self):
        """Get all collected data"""
        return self.stats.copy()

    def clear(self):
        """Reset all data"""
        self.stats.clear()
        self.file_history.clear()



import os.path
from collections import Counter
import inspect
import datetime
import uuid

import json
import sys
import os
import shutil
import sysconfig
import reprlib

import hashlib
import pickle

from praetor.transform_output import create_full_json
from praetor.process_monitor import DynamicProcessMonitor

custom_repr = reprlib.Repr()
custom_repr.maxlist = 80
custom_repr.maxdict = 40
custom_repr.maxstring = 1024
custom_repr.maxlevel = 3  # nesting depth


def get_caller_script_name():
    top_script_path = getattr(sys.modules['__main__'], '__file__', None)
    if top_script_path is not None:
        top_script_name = os.path.basename(top_script_path)
        return "".join(top_script_name.split('.')[:-1])
    else:
        return "python"


def generate_pipeline_id():
    """
    Creates a unique id for a pipeline run which includes the praetor version
    :return: Unique id
    """
    pipeline_id = "{}_provenance_{}".format(get_caller_script_name(), uuid.uuid4())
    return pipeline_id

def get_modules(pipeline_id, out_directory, modules):
    '''
    Function to create the agent_json.json file, including creating a unique id for the agent, determining all modules
    imported, find their versions, structure and input all infroamtion into agent_json.json
    :return: agent_json.json
    '''
    bindings = {
        "context": {"xsd": "http://www.w3.org/2001/XMLSchema#", "lsim": "http://example.org/", "urn_uuid": "urn:uuid:",
                    "run": "http://example.org/", "pyth" : "https://pypi.org/project"}}


    json_total = {}
    py_version = sys.version.split()[0]
    modules_pre = {f"pyth:{key}": value for key, value in modules.items()}
    bindings['var'] = {}
    bindings['var'].update(modules_pre)
    bindings['var']['python_version'] = py_version
    bindings['var']['lifeline'] = 'urn_uuid:{}'.format(pipeline_id)
    bindings['var']['run_cmd'] = f"{sys.executable} {' '.join(sys.argv)}"

    json_dir = out_directory + '/json/'

    json_total['agent'] = bindings

    agent_json = json_dir + "agent_json.json"

    with open(agent_json, 'w') as f:
        json.dump(json_total, f)

    return agent_json


def get_stdlib_modules():
    stdlib_path = sysconfig.get_paths().get('stdlib')
    module_names = set()

    for root, dirs, files in os.walk(stdlib_path):
        if '__pycache__' in dirs:
            dirs.remove('__pycache__')

        for filename in files:
            if filename.endswith('.py'):
                mod_name = filename[:-3]
                if mod_name == '__init__':
                    continue
                rel_dir = os.path.relpath(root, stdlib_path)
                if rel_dir == '.':
                    full_mod_name = mod_name
                else:
                    full_mod_name = rel_dir.replace(os.sep, '.') + '.' + mod_name
                module_names.add(full_mod_name)

        for dirname in dirs:
            if os.path.isfile(os.path.join(root, dirname, '__init__.py')):
                rel_dir = os.path.relpath(root, stdlib_path)
                if rel_dir == '.':
                    full_pkg_name = dirname
                else:
                    full_pkg_name = rel_dir.replace(os.sep, '.') + '.' + dirname
                module_names.add(full_pkg_name)
    return module_names


def get_non_builtin_modules_versions():
    builtin_modules = set(sys.builtin_module_names)
    stdlib_modules = get_stdlib_modules()
    modules_versions = {}

    mod_dict = dict(sys.modules)

    for name, module in mod_dict.items():
        if (module is not None and
            name not in builtin_modules and
            name not in stdlib_modules and
            not name.startswith('_') and
            '.' not in name and name != "praetor"):  # exclude submodules
            version = getattr(module, '__version__', None)
            if version is None:
                version = getattr(module, 'VERSION', None)
                if isinstance(version, tuple):
                    version = '.'.join(map(str, version))
            modules_versions[name] = version
    return modules_versions



class CallTracer:

    close_file_var = True
    out_handle = None
    agent_json = None

    def __init__(self, output_directory="./output/", block_list_mod=None, block_list_func=None, cpython=False,
                 bootstrap=False, store_large_values=False, only_main=False, slim=False, monitor_interval=1.0,
                 process_monitor=False):
        """Setup for the CallTracer class
        :param output_directory: Where generated provenance will be stored
        :param block_list_mod: block_list of python modules
        :param block_list_func: block_list of python functions
        :param cpython: Whether to record provenance for cpython functions
        :param bootstrap: Whether to record provenance for bootstrapped functions
        :param store_large_values: Whether to store copies of large values as files"""

        self.record_prov = True
        self.process_monitor = process_monitor
        if self.process_monitor:
            self.monitor = DynamicProcessMonitor(base_interval=monitor_interval)

        self.calls = {}
        self.session_id = generate_pipeline_id()  # change praetor to name of pipeline
        self.prov_id_counter = Counter()
        self.prov_id_cache = dict()

        self.store_large_values = store_large_values
        self.prefix = "run:"
        self.dtypes = {'int': 'int', 'unsignedInt': 'unsignedInt', 'hexBinary': 'hexBinary', 'NOTATION': 'NOTATION',
                  'nonPositiveInteger': 'nonPositiveInteger', 'float': 'float', 'ENTITY': 'ENTITY', 'bool': 'boolean',
                  'positiveInteger': 'positiveInteger', 'duration': 'duration', 'IDREFS': 'IDREFS',
                  'unsignedLong': 'unsignedLong', 'normalizedString': 'normalizedString',
                  'dateTimeStamp': 'dateTimeStamp',
                  'NMTOKEN': 'NMTOKEN', 'negativeInteger': 'negativeInteger', 'base64Binary': 'base64Binary',
                  'long': 'long', 'unsignedShort': 'unsignedShort', 'ENTITIES': 'ENTITIES', 'anyURI': 'anyURI',
                  'NMTOKENS': 'NMTOKENS', 'IDREF': 'IDREF', 'unsignedByte': 'unsignedByte', 'Name': 'Name',
                  'dayTimeDuration': 'dayTimeDuration', 'date': 'date', 'integer': 'integer', 'byte': 'byte',
                  'ID': 'ID',
                  'gMonth': 'gMonth', 'short': 'short', 'language': 'language', 'gMonthDay': 'gMonthDay',
                  'double': 'double', 'Decimal': 'decimal', 'gDay': 'gDay', 'gYearMonth': 'gYearMonth',
                  'QName': 'QName', 'datetime': 'dateTime', 'nonNegativeInteger': 'nonNegativeInteger',
                  'gYear': 'gYear',
                  'token': 'token', 'time': 'time', 'yearMonthDuration': 'yearMonthDuration', 'NCName': 'NCName',
                  'str': 'string'}
        self.out_directory = output_directory
        if not self.out_directory.endswith("/"):
            self.out_directory += "/"

        os.makedirs(self.out_directory + "json/", exist_ok=True)
        os.makedirs(self.out_directory + "big_entities/", exist_ok=True)
        self.out_handle = open(self.out_directory + "json/" + self.session_id + ".json", "a")

        script_path = os.path.abspath(__file__)
        script_name = script_path.split("/")[-1]
        shutil.copy(script_path, self.out_directory + "{}_{}".format(self.session_id, script_name))

        mods = get_non_builtin_modules_versions()
        self.agent_json = get_modules(self.session_id, self.out_directory, mods)

        self.last_activity = {"id": None, "end": None, "start": None, "name": None}
        self.activity_counter = {}
        self.bindings = {}

        self.block_list_modules = block_list_mod
        self.block_list_func = block_list_func
        self.cpython = cpython
        self.bootstrap = bootstrap
        self.only_main = only_main
        self.slim = slim
        self.prefixes = ("_", "<")

    def __call__(self, frame, event, arg):
        """Method to record metadata for each python call event
        :param frame: Python frame
        :param event: Type of frame
        :param arg: Argument of frame
        :returns self"""

        if event in ["call", "return"]:
            code = frame.f_code
            func_name = code.co_name
            module_name = frame.f_globals.get("__name__", None)

            if self.only_main and module_name != "__main__":
                return self

            if self.slim:
                if func_name.startswith(self.prefixes):
                    return self
                if "._" in module_name:
                    return self

            if module_name == "praetor":
                # print("passing my mod")
                return self

            if "praetor" in module_name:
                # print("passing module")
                return self
            # #
            # Ignore module-level frames
            if func_name == "<module>":
                return self

            if not self.record_prov:
                return self

            if self.bootstrap and module_name in ["importlib._bootstrap_external", "importlib._bootstrap"]:
                return self

            if self.block_list_modules:
                if module_name in self.block_list_modules:
                    return self
                for mod in self.block_list_modules:
                    if module_name.startswith(mod):
                        return self


            if self.block_list_func:
                if func_name in self.block_list_func:
                    return self

            argcount = code.co_argcount
            varnames = code.co_varnames

            if len(varnames) > 0:
                if self.slim and varnames[0] in ["cls", "self"]:
                    return self

            inputs = {
                varnames[i]: frame.f_locals.get(varnames[i])
                for i in range(argcount)
            }

            self.name = func_name
            self.module_name = module_name
            self.stack_id = str(id(frame))


            self.inputs = inputs

            if self.process_monitor:
                process_stats = self.monitor.high_freq_snapshot(func_name)
                self.process_count = process_stats["process_count"]
                self.total_memory = process_stats["total_rss_mb"]
                self.files_opened = process_stats["newly_opened_files"]

            if event == "call":

                self.start_time = self.date_time_stamp()
                self.prov_call_in()
                self.dump_json(mode="call")
                return self

            elif event == "return":
                self.output = arg
                self.end_time = self.date_time_stamp()
                self.prov_call_out()
                self.dump_json(mode="return")
                return self

        elif event in ["c_call", "c_return"]:

            if not self.cpython:
                return self

            cfunc = arg
            key = str(id(cfunc))
            func_name = getattr(cfunc, "__name__", None)
            module_name = getattr(cfunc, "__module__", None)

            self.name = func_name
            self.module_name = module_name
            self.stack_id = key

            code = frame.f_code
            argcount = code.co_argcount
            varnames = code.co_varnames

            if len(varnames) > 0:
                if self.slim and varnames[0] in ["cls", "self"]:
                    return self

            inputs = {
                varnames[i]: frame.f_locals.get(varnames[i])
                for i in range(argcount)
            }
            self.inputs = inputs

            if self.process_monitor:
                process_stats = self.monitor.high_freq_snapshot()
                self.process_count = process_stats["process_count"]
                self.total_memory = process_stats["total_rss_mb"]
                self.files_opened = process_stats["newly_opened_files"]

            if self.cpython and event == "c_call":
                self.start_time = self.date_time_stamp()
                self.prov_call_in()
                self.dump_json(mode="call")
                return self

            if self.cpython and event == "c_return":
                self.end_time = self.date_time_stamp()
                self.output = "None"
                self.prov_call_out()
                self.dump_json(mode="return")
                return self

        # code for if events are to be added
        # elif event == "exception":
        #     meta = self.calls.get(id(frame))
        #     if meta is not None:
        #         # arg is (exc_type, exc_value, traceback)
        #         exc_type, exc_value, _ = arg
        #         meta["exception"] = exc_value
        #     return self

        return self

    @staticmethod
    def date_time_stamp():
        """Generate current date and time stamp
        :returns: date and time stamp """
        date_time = datetime.datetime.now(datetime.timezone.utc).strftime('%Y-%m-%dT%H:%M:%S.%f')
        return date_time

    @staticmethod
    def remove_quotes_from_string(in_string):
        """Remove quotes from a string
        :param in_string: input string
        :returns in_string without quotes (out_string)"""
        out_string = [x for x in in_string if x not in ['"', "'"]]
        out_string = ''.join(out_string)
        return out_string

    @staticmethod
    def generate_persistent_id(obj):
        """Generate persistent id for python objects so that they are consistent across runs
        :param obj: python object
        :return: persistent id """
        try:
            # Try to serialize to JSON, for JSON serializable objects
            serialized = json.dumps(obj, sort_keys=True).encode('utf-8')
        except (TypeError, OverflowError):
            # Fallback: use pickle for other Python objects
            try:
                serialized = pickle.dumps(obj)
            except (pickle.PicklingError, TypeError, AttributeError): # just except all?
                return str(id(obj))
        # Generate SHA-256 hash of serialized representation
        persistent_id = hashlib.sha256(serialized).hexdigest()
        return persistent_id

    def gen_identifier(self, variable, naming_template="entity"):
        """
        General utility function to create prov ids for python objects
        :param variable: Value of object to create id for
        :param naming_template: Name related to the object in question to include in the id
        :return: ID for target object
        """
        try:
            prov_id = self.prov_id_cache[self.generate_persistent_id(variable)] # change back to id if too slow
        except KeyError:
            self.prov_id_counter[naming_template] += 1
            prov_id = '{}_{}_{}'.format(naming_template, self.session_id, self.prov_id_counter[naming_template])
            self.prov_id_cache[self.generate_persistent_id(variable)] = prov_id # change back to id if too slow

        return prov_id

    def find_type(self, value, value_id):
        """Determine the type of any input python object
        :param value: Input python object
        :param value_id: ID of the input python object
        :return: Type and truncated string representation of the input python object
        """
        # mod_list = self.praetor_settings_dict["modules"] # need to get from some place, probably enviro var
        if callable(value) or inspect.ismodule(value) or inspect.isclass(value):#  or type(value).__module__ in mod_list:
            name = getattr(value, '__name__', None)
            out_value = f"<callable {name}>"
            prov_type = 'string'
        else:
            out_value = custom_repr.repr(value) # truncating the value, need to not do that
            try:
                out_value_full = json.dumps(value, ensure_ascii=False)
            except (TypeError, OverflowError):
                out_value_full = out_value
            out_value = self.remove_quotes_from_string(out_value)
            # print(out_value, out_value_full)
            if self.store_large_values:
                if len(out_value_full) > 1024:
                    self.large_object_handling(out_value_full, value_id)
            try:
                out_type = type(value).__name__
                if out_type in self.dtypes.keys():
                    prov_type = self.dtypes[out_type]
                else:
                    prov_type = 'string'
            except:
                prov_type = 'string'
        return out_value, prov_type

    def large_object_handling(self, value, object_id):
        """Save large entities to file to save on provenance size
        :param value: Value of object to save
        :param object_id: ID of object to save
        :return: Name of file which was dumped to
        """
        file_name = self.out_directory + "/big_entities/{}.json".format(object_id.split(":")[1])

        if not os.path.isfile(file_name):
            with open(file_name, "w") as f:
                json.dump(value, f)

        return file_name

    def track_call(self):
        """Trace the python stack to determine if a function was called by another"""
        stack = inspect.stack()
        caller = 'main'
        for frame in stack:
            if frame.function == "<module>":
                break
            elif frame.function == self.last_activity['name']:
                caller = self.last_activity['name']
                break
        return caller

    def start_monitoring(self):
        """Start continuous background monitoring"""
        self.monitor.start()

    def stop_monitoring(self):
        """Stop continuous background monitoring"""
        self.monitor.stop()

    def get_full_trace(self):
        """Get all trace data with memory snapshots"""
        return self.monitor.get_stats()

    def clear(self):
        """Reset all data"""
        self.metadata.clear()
        self.monitor.clear()

    def prov_call_in(self):
        """Format call metadata into json format provenance"""
        self.bindings['{}'.format(self.stack_id)] = {}
        self.bindings['{}'.format(self.stack_id)]['messageStartTime'] = {"@type": "xsd:dateTime", "@value": self.start_time}
        self.bindings['{}'.format(self.stack_id)]['moduleName'] = {"@type": "xsd:string", "@value": self.module_name}
        self.bindings['{}'.format(self.stack_id)]['activityName'] = {"@type": "xsd:string", "@value": self.name}
        self.bindings['{}'.format(self.stack_id)]['message'] = {"@id": "urn_uuid:{}_{}".format(self.session_id, self.stack_id)}
        if self.process_monitor:
            self.bindings['{}'.format(self.stack_id)]['memory_call'] = {"@type": "xsd:float", "@value": self.total_memory}

        # add gate to see if it is on the stack
        # stack_function = self.track_call()
        if self.last_activity['id']:
            self.bindings['{}'.format(self.stack_id)]['message2'] = {"@id": "urn_uuid:{}_{}".format(self.session_id, self.last_activity['id'])}
            self.bindings['{}'.format(self.stack_id)]['message2StartTime'] = {"@type": "xsd:dateTime", "@value":self.last_activity['end']}
            self.bindings['{}'.format(self.stack_id)]['message2EndTime'] = {"@type": "xsd:dateTime", "@value": self.last_activity['start']}


        counter = 0
        for key, value in self.inputs.items():
            in_id = self.generate_persistent_id(value)
            in_value, in_type = self.find_type(value, self.prefix + in_id)
            self.bindings['{}'.format(self.stack_id)]['input_{}'.format(counter)] = {'@id' : self.prefix + in_id, '@value': in_value,
                                                                     '@type': "xsd:{}".format(in_type), '@role': key}
            counter += 1

        self.last_activity['id'] = self.stack_id
        self.last_activity['name'] = self.name
        self.last_activity['end'] = self.start_time
        self.last_activity['start'] = self.start_time


    def prov_call_out(self):
        """Format return metadata into json format provenance"""
        self.bindings['{}'.format(self.stack_id)] = {}
        self.bindings['{}'.format(self.stack_id)]['messageEndTime'] = {"@type": "xsd:dateTime", "@value": self.end_time}
        self.bindings['{}'.format(self.stack_id)]['moduleName'] = {"@type": "xsd:string",
                                                                   "@value": self.module_name}
        self.bindings['{}'.format(self.stack_id)]['activityName'] = {"@type": "xsd:string", "@value": self.name}
        self.bindings['{}'.format(self.stack_id)]['message'] = {
            "@id": "urn_uuid:{}_{}".format(self.session_id, self.stack_id)}
        if self.process_monitor:
            self.bindings['{}'.format(self.stack_id)]['memory_return'] = {"@type": "xsd:float", "@value": self.total_memory}
            if len(self.files_opened) > 0:
                self.bindings['{}'.format(self.stack_id)]['file_access'] = {"@type": "xsd:string", "@value": self.files_opened}

        counter = 0
        for key, value in self.inputs.items():
            in_id = self.generate_persistent_id(value)
            in_value, in_type = self.find_type(value, self.prefix + in_id)
            self.bindings['{}'.format(self.stack_id)]['input_{}'.format(counter)] = {'@id': self.prefix + in_id,
                                                                                     '@value': in_value,
                                                                                     '@type': "xsd:{}".format(
                                                                                         in_type), '@role': key}
            counter += 1

        output_list = [self.output]
        if output_list:
            for i, output_item in enumerate(output_list):
                out_id = self.generate_persistent_id(output_item)
                out_value, out_type = self.find_type(output_item, self.prefix + out_id)
                self.bindings['{}'.format(self.stack_id)]['output_{}'.format(i)] = {
                    '@id': self.prefix + out_id, '@value': out_value,
                    '@type': "xsd:{}".format(out_type)}

        if self.last_activity['id']:
            self.bindings['{}'.format(self.stack_id)]['message2'] = {"@id": "urn_uuid:{}_{}".format(self.session_id, self.last_activity['id'])}
            self.bindings['{}'.format(self.stack_id)]['message2StartTime'] = {"@type": "xsd:dateTime", "@value":self.last_activity['end']}
            self.bindings['{}'.format(self.stack_id)]['message2EndTime'] = {"@type": "xsd:dateTime", "@value": self.last_activity['start']}


        self.last_activity['id'] = self.stack_id
        self.last_activity['name'] = self.name
        try:
            self.last_activity['end'] = self.end_time
        except AttributeError:
            self.last_activity['end'] = "-"
        try:
            self.last_activity['start'] = self.start_time
        except AttributeError:
            self.last_activity['start'] = "-"

    def dump_json(self, mode):
        """dump json provenance to file
        :param mode: call or return"""
        json_metadata = self.bindings.pop(self.stack_id, None)
        # print(json_metadata)
        new_json = {'@id': '{}'.format(self.stack_id), '@mode': mode, '@data': json_metadata}
        json_str = json.dumps(new_json)
        self.out_handle.write(json_str + '\n')
        self.out_handle.flush()


    def close(self):
        """Close json dump document at end of provenance generation"""
        if self.close_file_var and self.out_handle:
            self.stop_monitoring()
            self.record_prov = False
            with open(self.out_directory + "stats.txt", "w") as f:
                f.write(str(self.monitor.get_stats()))
            create_full_json(self.agent_json, self.out_handle.name)
            self.out_handle.close()
            self.close_file_var = False




from collections import Counter
from datetime import datetime
import logging
import os
import requests
import subprocess


import pandas as pd
from pandas import json_normalize

# add function to remove prefixes from results

DATABASE_HOST_URL = os.environ.get('DATABASE_HOST_URL', 'http://127.0.0.1:3030/')
REPOSITORY_ID = os.environ.get('REPOSITORY_ID', 'ds')
SPARQL_REST_URL = DATABASE_HOST_URL + REPOSITORY_ID

prefixes = '''
PREFIX prov: <http://www.w3.org/ns/prov#>
PREFIX run: <http://example.org/>
PREFIX urn_uuid: <urn:uuid:>
PREFIX ivoa: <https://www.ivoa.net/documents/ProvenanceDM/20200411/Provenance.html>
PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
PREFIX rdf: <http://www.w3.org/1999/02/22-rdf-syntax-ns#>
PREFIX xsd: <http://www.w3.org/2001/XMLSchema#>
PREFIX var: <http://openprovenance.org/var#>
PREFIX prtr: <http://example.org/> 
PREFIX pyth: <https://pypi.org/project/> 
'''

prefix_dict = {'http://www.w3.org/ns/prov#': 'prov:',
               'http://example.org/': 'run:',
               'https://praetor.pages.mpcdf.de/praetor_provenance/': 'prtr:',
               'http://www.w3.org/2000/01/rdf-schema#': 'rdfs:'}

def clear_response(arr):
    logging.debug('StaticUtils: clearing given array')
    unwanted_strings = ['s', '', ' ', 's,p', 's,p,o', 'p,o', 'o', 'p', 'g', 'g,s,p,o', 's,p,o,r',
                        'activity,entity,p,dt,value,role,startTime,endTime', 's,s_count', 's,s_count,graphs', 's_count',
                        's,s_count,p,o,o_count', 'uuid,s,p,o', 's,g,s_count', 's,s_count,roles', 's,ts,te', 's,u']

    for string in unwanted_strings:
        if string in arr:
            arr.remove(string)

    logging.debug('StaticUtils: cleared list: ' + str(arr))

    return arr


def query_handler(query):
    '''
    Sends query to triple store
    :param query:
    :return: response of query
    '''
    response = requests.post(SPARQL_REST_URL,
                             data={'query': query},
                             timeout=86400)

    return response


def upload_provenance(file_name):
    '''
    Function for uploading provenance files to the database
    :param file_name: name of provenance file to upload
    :return: Name of the graph in the database
    '''
    file_name_short = file_name.split('/')[-1]
    pipeline = 'http://' + file_name_short.replace('.ttl', '')
    data = open(file_name).read()
    headers = {'Content-Type': 'text/turtle;charset=utf-8'}
    url = SPARQL_REST_URL + '/data'
    requests.post(url, params={'graph': pipeline}, data=data, headers=headers)
    return pipeline

def convert_to_datetime_exception(datetime_str):
    '''
    Converts a string representation of datetime to a datetime object - temperamental
    :param datetime_str:
    :return: datetime object
    '''
    try:
        time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%f')
    except ValueError:
        try:
            time = datetime.strptime(datetime_str[:-6], '%Y-%m-%dT%H:%M:%S.%f')
        except ValueError:
            try:
                time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S.%fZ')
            except ValueError:
                try:
                    time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%S')
                except:
                    time = datetime.strptime(datetime_str, '%Y-%m-%dT%H:%M:%SZ')


    return time


def result_to_df(response, pre_dict=prefix_dict):
    '''
    Converts a response from the database to a pandas dataframe
    :param response: response to convert
    :return: pandas dataframe format response
    '''
    df = json_normalize(response.json()['results']['bindings'], meta='value')
    cols = [c for c in df.columns if c[-6:] == '.value']
    df2 = df[cols]
    col_names = [x for x in cols]

    if 'startTime.value' in col_names and 'endTime.value' in col_names:
        start_strs = df['startTime.value'].tolist()
        end_strs = df['endTime.value'].tolist()

        start_str = [x for x in start_strs]
        datetime_start = [convert_to_datetime_exception(x) for x in start_str]

        end_str = [x for x in end_strs]
        datetime_end = [convert_to_datetime_exception(x) for x in end_str]

        durations = [(x - y).total_seconds() for x, y in zip(datetime_end, datetime_start)]

        df2.insert(1, 'duration (s)', durations)

    df2 = df2.drop_duplicates()
    for key, value in pre_dict.items():
        df2 = df2.replace(key, value, regex=True)
    return df2


def user_defined_query(query):
    '''
    Function to pass a SPARQL query to the database
    :param query: string SPARQL query
    :return: dataframe of the results
    '''
    res = query_handler(query)
    response_as_string = clear_response(res.text.split("\r\n"))
    # print(response_as_string)
    df = result_to_df(res)
    return df

def modules_query(prov_name, prefixes=prefixes):
    '''
    Function to return module names and versions
    :param prov_name: name of provenance graph
    :return:
    '''

    query = prefixes + '''
SELECT ?modname ?modver
FROM NAMED <''' + prov_name + '''>
WHERE {
GRAPH ?g {

?agentID a prov:Agent ;
    ?modname ?modver .

}
}
'''
    res = query_handler(query)
    df = result_to_df(res)
    return df


def track_functions(prov_name, input_name=None, output_name=None, function_id=None, trace_back=True, prefixes=prefixes):
    '''
    Determine what functions have executed before or after a given input value, output value, or function id.
    Note, values may appear more than once, in this case the first instance is always used.
    :param prov_name:
    :param input_name:
    :param output_name:
    :param function_id:
    :param trace_back:
    :param prefixes:
    :return:
    '''
    if input_name:
        find_ba = prefixes + '''
        SELECT ?actID ?time
        FROM NAMED <''' + prov_name + '''>

        WHERE {
        GRAPH ?g {

        ?entityID prov:value "''' + input_name + '''" .
        ?be prov:entity ?entityID .
        ?actID prov:qualifiedUsage ?be .
        ?ba prov:hadActivity ?actID .
        ?actID prov:startedAtTime ?time .
        }
        }
        '''
        ba = user_defined_query(find_ba)
        time = ba['time.value'][0]
    elif output_name:
        find_ba = prefixes + '''
        SELECT ?actID ?time
        FROM NAMED <''' + prov_name + '''>

        WHERE {
        GRAPH ?g {

        ?entityID prov:value "''' + output_name + '''" .
        ?entityID prov:wasGeneratedBy ?actID .
        ?ba prov:hadActivity ?actID .
        ?actID prov:startedAtTime ?time .
        }
        }
        '''
        ba = user_defined_query(find_ba)
        time = ba['time.value'][0]
    elif function_id:
        find_ba = prefixes + '''
        SELECT ?time
        FROM NAMED <''' + prov_name + '''>

        WHERE {
        GRAPH ?g {
        <''' + function_id + '''> prov:startedAtTime ?time .
        }
        }
        '''
        ba = user_defined_query(find_ba)
        time = ba['time.value'][0]
    else:
        raise Exception("Input error, please enter a value for either the input_name, output_name, or function_id")

    date_time_mine = pd.to_datetime(time)

    get_all_timing = prefixes + '''
    SELECT ?start ?name
    FROM NAMED <''' + prov_name + '''>
    WHERE {
    GRAPH ?g{

    ?actID a prov:Activity .
    ?actID prov:startedAtTime ?start.
    ?actID prtr:activityName ?name .
    }
    }
    '''

    all_times = user_defined_query(get_all_timing)
    all_times['start.value'] = pd.to_datetime(all_times['start.value'])
    if trace_back:
        df_filtered = all_times[all_times['start.value'] <= date_time_mine]
    else:
        df_filtered = all_times[all_times['start.value'] >= date_time_mine]

    return df_filtered

def function_query(prov_name, prefixes=prefixes, group_by=None):
    """

    :param prov_name:
    :param prefixes:
    :param group_by:
    :return:
    """

    query = prefixes + '''
    SELECT ?funcID ?funcName ?inputNames ?inputValues ?outputValues 

    FROM NAMED <''' + prov_name + '''>
    WHERE {
    GRAPH ?g {
    
    ?funcID a prov:Activity;
        prtr:activityName ?funcName .
        
    OPTIONAL {
    ?b prov:entity ?inputID .
    ?funcID prov:qualifiedUsage ?b .
    ?b prov:hadRole ?inputNames .
    ?inputID prov:value ?inputValues .
    }
    
    OPTIONAL {
    ?outputID prov:wasGeneratedBy ?funcID .
    ?outputID prov:value ?outputValues . 
    }
    }
    }
    '''

    full_data_frame = user_defined_query(query)

    if group_by:
        groups = full_data_frame[full_data_frame["funcName.value"] == group_by]
    else:
        groups = full_data_frame.sort_values(by="funcName.value", ascending=False)

    return groups


def module_query(prov_name, module_name, prefixes=prefixes):
    """

    :param prov_name:
    :param module_name:
    :param prefixes:
    :return:
    """

    query = prefixes + '''
    SELECT ?funcName ?funcID ?start ?end 

    FROM NAMED <''' + prov_name + '''> 
    WHERE {
    GRAPH ?g {

    ?funcID a prov:Activity ;
        prtr:activityName ?funcName ;
        prtr:activitySource "''' + module_name + '''" .

    OPTIONAL {
    ?funcID prov:startedAtTime ?start .
    }

    OPTIONAL {
    ?funcID prov:endedAtTime ?end .
    }

    }
    }

    '''
    full_data_frame = user_defined_query(query)
    return full_data_frame


def duration_query(prov_name, prefixes=prefixes):
    query = prefixes + '''
    SELECT ?start ?end ?funcName ?module

    FROM NAMED <''' + prov_name + '''>
    WHERE {
    GRAPH ?g {

    ?funcID a prov:Activity ;
            prtr:activityName ?funcName ;
            prtr:activitySource ?module .


        ?funcID prov:startedAtTime ?start .



        ?funcID prov:endedAtTime ?end .


        }
        }
    '''
    df = user_defined_query(query)
    duration = df[['start.value', 'end.value']] = df[['start.value', 'end.value']].apply(pd.to_datetime)
    df['time_diff'] = df['end.value'] - df['start.value']
    df["diff_sec"] = df["time_diff"] / pd.Timedelta(seconds=1)
    return df
